import pytest
from pytest_mock import MockerFixture

from by_config.config import Config


class TestConfig:
    def test_is_singleton(self) -> None:
        config1: Config = Config()
        config2: Config = Config()
        assert config1 is config2

    @pytest.fixture
    def config(self) -> Config:
        config: Config = Config()
        return config.empty()

    def test_load_empty_path(self, config: Config) -> None:
        config.load()
        assert config.payload == {}

    def test_load_specific_file_path(
        self, config: Config, mocker: MockerFixture
    ) -> None:
        expected = {"key": "value"}
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch(
            "by_config.loader.YamlLoader.load",
            return_value={"key": "value", "use": True},
        )
        config.load("path/to/file.yaml")
        assert config.payload == expected

    def test_load_multiple_file_paths(
        self, config: Config, mocker: MockerFixture
    ) -> None:
        expected = {"key1": "value1", "key2": "value2"}
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch(
            "by_config.loader.YamlLoader.load",
            side_effect=[
                {"key1": "value1", "use": True},
                {"key2": "value2", "use": True},
            ],
        )
        config.load("path/to/file1.yaml", "path/to/file2.yaml")
        assert config.payload == expected

    def test_load_dir(self, config: Config, mocker: MockerFixture) -> None:
        path: str = "dir/to/load"
        expected = {"key": "value"}
        mocker.patch("os.path.isfile", return_value=False)
        mocker.patch("os.path.isdir", return_value=True)
        mocker.patch("os.walk", return_value=[("path/to/dir", [], ["file.yaml"])])
        mocker.patch(
            "by_config.loader.YamlLoader.load",
            return_value={"key": "value", "use": True},
        )
        config.load(path)
        assert config.payload == expected

    def test_load_file_with_use(self, config: Config, mocker: MockerFixture) -> None:
        paths: list[str] = ["path/to/file1.yaml", "path/to/file2.yaml"]
        expected = {"key1": "value1"}
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch(
            "by_config.loader.YamlLoader.load",
            side_effect=[
                {"key1": "value1", "use": True},
                {"key2": "value2"},
            ],
        )
        config.load(*paths)
        assert config.payload == expected

    def test_load_file_with_priority(
        self, config: Config, mocker: MockerFixture
    ) -> None:
        paths: list[str] = ["path/to/file1.yaml", "path/to/file2.yaml"]
        expected = {"key": "value1"}
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch(
            "by_config.loader.YamlLoader.load",
            side_effect=[
                {"key": "value1", "priority": 1, "use": True},
                {"key": "value2", "use": True},
            ],
        )
        config.load(*paths)
        assert config.payload == expected

    def test_get_with_str_path(self, config: Config) -> None:
        config.payload = {"key": "value"}
        assert config.get("key") == "value"

    def test_get_with_str_path_not_found(self, config: Config) -> None:
        config.payload = {"key": "value"}
        assert config.get("non_existent_key") is None

    def test_get_with_deep_path(self, config: Config) -> None:
        config.payload = {"key": {"sub_key": "value"}}
        assert config.get("key.sub_key") == "value"

    def test_get_with_list_path(self, config: Config) -> None:
        config.payload = {"key": {"sub_key": "value"}}
        assert config.get(["key", "sub_key"]) == "value"



import pytest
from pytest_mock import MockerFixture

from by_config.loader import YamlLoader


class TestYamlLoader:
    def test_load(self) -> None:
        YamlLoader()
        assert True

    @pytest.fixture
    def loader(self) -> YamlLoader:
        return YamlLoader()

    def test_load_file(self, loader: YamlLoader, mocker: MockerFixture) -> None:
        path: str = "a.yaml"
        expected: dict = {"a": "b"}
        mocker.patch("builtins.open", mocker.mock_open(read_data="a: b"))
        mocker.patch("os.path.exists", return_value=True)
        mocker.patch("os.path.isfile", return_value=True)

        assert loader.load(path) == expected

    def test_load_file_not_exists(
        self, loader: YamlLoader, mocker: MockerFixture
    ) -> None:
        path: str = "a.yaml"
        expected: dict = {}
        mocker.patch("os.path.exists", return_value=False)
        mocker.patch("builtins.open", mocker.mock_open(read_data="a: b"))

        assert loader.load(path) == expected

    def test_load_path_not_file(
        self, loader: YamlLoader, mocker: MockerFixture
    ) -> None:
        path: str = "a.yaml"
        expected: dict = {}
        mocker.patch("os.path.exists", return_value=False)
        mocker.patch("builtins.open", mocker.mock_open(read_data="a: b"))
        assert loader.load(path) == expected

    def test_load_path_not_valid(
        self, loader: YamlLoader, mocker: MockerFixture
    ) -> None:
        path: str = "a.txt"
        expected: dict = {}
        mocker.patch("os.path.exists", return_value=True)
        mocker.patch("os.path.isfile", return_value=True)
        mocker.patch("builtins.open", mocker.mock_open(read_data="a: b"))
        assert loader.load(path) == expected



import os
import typing as t
import threading

from by_config.loader import YamlLoader


class Config:
    _instance: "Config" = None
    _lock: threading.Lock = threading.Lock()

    def __new__(cls):
        if cls._instance is None:
            with cls._lock:
                if cls._instance is None:
                    cls._instance = super().__new__(cls)
        return cls._instance

    def __init__(self) -> None:
        self.loaders = [YamlLoader()]
        self.payload: dict = {}
        self.filters = [self.is_use]

    def _merge_deep(self, a: dict, b: dict) -> None:
        for key, value in b.items():
            if isinstance(value, dict) and key in a:
                self._merge_deep(a[key], value)
            else:
                a[key] = value

    def empty(self) -> t.Self:
        self.payload = {}
        return self

    def is_use(self, item: dict) -> bool:
        return item.pop("use", False)

    def load(self, *paths: str, verbose: bool = False) -> None:
        paths = [path for path in paths if os.path.isfile(path)] + [
            os.path.join(root, file)
            for path in paths
            if os.path.isdir(path)
            for root, _, files in os.walk(path)
            for file in files
        ]

        raw_items = [loader.load(path, verbose) for path in paths for loader in self.loaders]

        for f in self.filters:
            raw_items = [item for item in raw_items if f(item)]

        raw_items = sorted(raw_items, key=lambda x: x.pop("priority", -1))

        for item in raw_items:
            self._merge_deep(self.payload, item)

    def get(self, path: str | list[str], default=None) -> t.Any:
        keys = path.split(".") if isinstance(path, str) else path
        ret: t.Any = self.payload
        for key in keys:
            if key not in ret:
                return default
            ret = ret[key]
        return ret if ret else default



import logging
import os
import yaml


logger = logging.getLogger("ConfigLoader")


class YamlLoader:
    def is_valid(self, path: str) -> bool:
        return all(
            (
                os.path.exists(path),
                os.path.isfile(path),
                path.lower().endswith(".yaml") or path.lower().endswith(".yml"),
            )
        )

    def load(self, path: str, verbose: bool = False) -> dict:
        path = os.path.abspath(path)
        if not self.is_valid(path):
            return {}

        if verbose:
            logger.info(f"Loading YAML file: {path}")

        with open(path, "r") as file:
            return yaml.safe_load(file)


class JsonLoader:
    pass
